@@ -1,81 +1,74 @@
 import logging
 import re
 import struct
+from collections import OrderedDict, defaultdict
+from concurrent.futures import Future
+from threading import Event
+from typing import List, NamedTuple, Optional, Union
+from urllib.parse import urlparse
 
-from collections import defaultdict, namedtuple, OrderedDict
+# noinspection PyPackageRequirements
 from Crypto.Cipher import AES
-from requests.exceptions import ChunkedEncodingError
+# noinspection PyPackageRequirements
+from Crypto.Util.Padding import unpad
+from requests import Response
+from requests.exceptions import ChunkedEncodingError, ConnectionError, ContentDecodingError
 
-from streamlink.compat import urlparse, str
 from streamlink.exceptions import StreamError
-from streamlink.stream import hls_playlist
 from streamlink.stream.ffmpegmux import FFMPEGMuxer, MuxedStream
+from streamlink.stream.hls_playlist import Key, M3U8, Map, Segment, load as load_hls_playlist
 from streamlink.stream.http import HTTPStream
-from streamlink.stream.segmented import (SegmentedStreamReader,
-                                         SegmentedStreamWriter,
-                                         SegmentedStreamWorker)
-from streamlink.utils import LazyFormatter
+from streamlink.stream.segmented import SegmentedStreamReader, SegmentedStreamWorker, SegmentedStreamWriter
+from streamlink.utils.cache import LRUCache
+from streamlink.utils.formatter import Formatter
 
 log = logging.getLogger(__name__)
-Sequence = namedtuple("Sequence", "num segment")
 
 
-def num_to_iv(n):
-    return struct.pack(">8xq", n)
-
-
-def pkcs7_decode(paddedData, keySize=16):
-    '''
-    Remove the PKCS#7 padding
-    '''
-    # Use ord + [-1:] to support both python 2 and 3
-    val = ord(paddedData[-1:])
-    if val > keySize:
-        raise StreamError("Input is not padded or padding is corrupt, got padding size of {0}".format(val))
-
-    return paddedData[:-val]
+class Sequence(NamedTuple):
+    num: int
+    segment: Segment
 
 
 class HLSStreamWriter(SegmentedStreamWriter):
-    def __init__(self, reader, *args, **kwargs):
-        options = reader.stream.session.options
-        kwargs["retries"] = options.get("hls-segment-attempts")
-        kwargs["threads"] = options.get("hls-segment-threads")
-        kwargs["timeout"] = options.get("hls-segment-timeout")
-        kwargs["ignore_names"] = options.get("hls-segment-ignore-names")
-        SegmentedStreamWriter.__init__(self, reader, *args, **kwargs)
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        options = self.session.options
 
         self.byterange_offsets = defaultdict(int)
+        self.map_cache: LRUCache[Sequence.segment.map.uri, Future] = LRUCache(self.threads)
         self.key_data = None
         self.key_uri = None
         self.key_uri_override = options.get("hls-segment-key-uri")
         self.stream_data = options.get("hls-segment-stream-data")
 
-        if self.ignore_names:
-            # creates a regex from a list of segment names,
-            # this will be used to ignore segments.
-            self.ignore_names = list(set(self.ignore_names))
-            self.ignore_names = "|".join(list(map(re.escape, self.ignore_names)))
-            self.ignore_names_re = re.compile(r"(?:{blacklist})\.ts".format(
-                blacklist=self.ignore_names), re.IGNORECASE)
+        self.ignore_names = False
+        ignore_names = {*options.get("hls-segment-ignore-names")}
+        if ignore_names:
+            segments = "|".join(map(re.escape, ignore_names))
+            self.ignore_names = re.compile(rf"(?:{segments})\.ts", re.IGNORECASE)
 
-    def create_decryptor(self, key, sequence):
+    @staticmethod
+    def num_to_iv(n: int) -> bytes:
+        return struct.pack(">8xq", n)
+
+    def create_decryptor(self, key: Key, num: int) -> AES:
         if key.method != "AES-128":
-            raise StreamError("Unable to decrypt cipher {0}", key.method)
+            raise StreamError(f"Unable to decrypt cipher {key.method}")
 
         if not self.key_uri_override and not key.uri:
             raise StreamError("Missing URI to decryption key")
 
         if self.key_uri_override:
             p = urlparse(key.uri)
-            key_uri = LazyFormatter.format(
-                self.key_uri_override,
-                url=key.uri,
-                scheme=p.scheme,
-                netloc=p.netloc,
-                path=p.path,
-                query=p.query,
-            )
+            formatter = Formatter({
+                "url": lambda: key.uri,
+                "scheme": lambda: p.scheme,
+                "netloc": lambda: p.netloc,
+                "path": lambda: p.path,
+                "query": lambda: p.query,
+            })
+            key_uri = formatter.format(self.key_uri_override)
         else:
             key_uri = key.uri
 
@@ -87,96 +80,141 @@ class HLSStreamWriter(SegmentedStreamWriter):
             self.key_data = res.content
             self.key_uri = key_uri
 
-        iv = key.iv or num_to_iv(sequence)
+        iv = key.iv or self.num_to_iv(num)
 
         # Pad IV if needed
         iv = b"\x00" * (16 - len(iv)) + iv
 
         return AES.new(self.key_data, AES.MODE_CBC, iv)
 
-    def create_request_params(self, sequence):
+    def create_request_params(self, segment: Union[Segment, Map]):
         request_params = dict(self.reader.request_params)
         headers = request_params.pop("headers", {})
 
-        if sequence.segment.byterange:
-            bytes_start = self.byterange_offsets[sequence.segment.uri]
-            if sequence.segment.byterange.offset is not None:
-                bytes_start = sequence.segment.byterange.offset
+        if segment.byterange:
+            bytes_start = self.byterange_offsets[segment.uri]
+            if segment.byterange.offset is not None:
+                bytes_start = segment.byterange.offset
 
-            bytes_len = max(sequence.segment.byterange.range - 1, 0)
+            bytes_len = max(segment.byterange.range - 1, 0)
             bytes_end = bytes_start + bytes_len
-            headers["Range"] = "bytes={0}-{1}".format(bytes_start, bytes_end)
-            self.byterange_offsets[sequence.segment.uri] = bytes_end + 1
+            headers["Range"] = f"bytes={bytes_start}-{bytes_end}"
+            self.byterange_offsets[segment.uri] = bytes_end + 1
 
         request_params["headers"] = headers
 
         return request_params
 
-    def fetch(self, sequence, retries=None):
-        if self.closed or not retries:
+    def put(self, sequence: Sequence):
+        if self.closed:
             return
 
+        if sequence is None:
+            self.queue(None, None)
+        else:
+            # always queue the segment's map first if it exists
+            if sequence.segment.map is not None:
+                future = self.map_cache.get(sequence.segment.map.uri)
+                # use cached map request if not a stream discontinuity
+                # don't fetch multiple times when map request of previous segment is still pending
+                if future is None or sequence.segment.discontinuity:
+                    future = self.executor.submit(self.fetch_map, sequence)
+                    self.map_cache.set(sequence.segment.map.uri, future)
+                self.queue(sequence, future, True)
+
+            # regular segment request
+            future = self.executor.submit(self.fetch, sequence)
+            self.queue(sequence, future, False)
+
+    def fetch(self, sequence: Sequence) -> Optional[Response]:
         try:
-            request_params = self.create_request_params(sequence)
-            # skip ignored segment names
-            if self.ignore_names and self.ignore_names_re.search(sequence.segment.uri):
-                log.debug("Skipping segment {0}".format(sequence.num))
-                return
+            return self._fetch(sequence.segment, self.stream_data and not sequence.segment.key)
+        except StreamError as err:  # pragma: no cover
+            log.error(f"Failed to fetch segment {sequence.num}: {err}")
 
-            return self.session.http.get(sequence.segment.uri,
-                                         stream=(self.stream_data
-                                                 and not sequence.segment.key),
-                                         timeout=self.timeout,
-                                         exception=StreamError,
-                                         retries=self.retries,
-                                         **request_params)
-        except StreamError as err:
-            log.error("Failed to open segment {0}: {1}", sequence.num, err)
+    def fetch_map(self, sequence: Sequence) -> Optional[Response]:
+        try:
+            return self._fetch(sequence.segment.map, False)
+        except StreamError as err:  # pragma: no cover
+            log.error(f"Failed to fetch map for segment {sequence.num}: {err}")
+
+    def _fetch(self, segment: Union[Segment, Map], stream: bool) -> Optional[Response]:
+        if self.closed or not self.retries:  # pragma: no cover
             return
 
-    def write(self, sequence, res, chunk_size=8192):
+        request_params = self.create_request_params(segment)
+
+        return self.session.http.get(
+            segment.uri,
+            stream=stream,
+            timeout=self.timeout,
+            exception=StreamError,
+            retries=self.retries,
+            **request_params
+        )
+
+    def should_filter_sequence(self, sequence: Sequence) -> bool:
+        return self.ignore_names and self.ignore_names.search(sequence.segment.uri) is not None
+
+    def write(self, sequence: Sequence, *args, **kwargs):
+        if not self.should_filter_sequence(sequence):
+            try:
+                return self._write(sequence, *args, **kwargs)
+            finally:
+                # unblock reader thread after writing data to the buffer
+                if not self.reader.filter_event.is_set():
+                    log.info("Resuming stream output")
+                    self.reader.filter_event.set()
+
+        # block reader thread if filtering out segments
+        elif self.reader.filter_event.is_set():
+            log.info("Filtering out segments and pausing stream output")
+            self.reader.filter_event.clear()
+
+    def _write(self, sequence: Sequence, res: Response, is_map: bool):
         if sequence.segment.key and sequence.segment.key.method != "NONE":
             try:
-                decryptor = self.create_decryptor(sequence.segment.key,
-                                                  sequence.num)
+                decryptor = self.create_decryptor(sequence.segment.key, sequence.num)
             except StreamError as err:
-                log.error("Failed to create decryptor: {0}", err)
+                log.error(f"Failed to create decryptor: {err}")
                 self.close()
                 return
 
             data = res.content
             # If the input data is not a multiple of 16, cut off any garbage
-            garbage_len = len(data) % 16
+            garbage_len = len(data) % AES.block_size
             if garbage_len:
-                log.debug("Cutting off {0} bytes of garbage "
-                          "before decrypting", garbage_len)
+                log.debug(f"Cutting off {garbage_len} bytes of garbage before decrypting")
                 decrypted_chunk = decryptor.decrypt(data[:-garbage_len])
             else:
                 decrypted_chunk = decryptor.decrypt(data)
 
-            self.reader.buffer.write(pkcs7_decode(decrypted_chunk))
+            chunk = unpad(decrypted_chunk, AES.block_size, style="pkcs7")
+            self.reader.buffer.write(chunk)
         else:
             try:
-                for chunk in res.iter_content(chunk_size):
+                for chunk in res.iter_content(8192):
                     self.reader.buffer.write(chunk)
-            except ChunkedEncodingError:
-                log.error("Download of segment {0} failed", sequence.num)
-
+            except (ChunkedEncodingError, ContentDecodingError, ConnectionError) as err:
+                log.error(f"Download of segment {sequence.num} failed ({err})")
                 return
 
-        log.debug("Download of segment {0} complete", sequence.num)
+        if is_map:
+            log.debug(f"Segment initialization {sequence.num} complete")
+        else:
+            log.debug(f"Segment {sequence.num} complete")
 
 
 class HLSStreamWorker(SegmentedStreamWorker):
     def __init__(self, *args, **kwargs):
-        SegmentedStreamWorker.__init__(self, *args, **kwargs)
+        super().__init__(*args, **kwargs)
         self.stream = self.reader.stream
 
         self.playlist_changed = False
-        self.playlist_end = None
-        self.playlist_sequence = -1
-        self.playlist_sequences = []
-        self.playlist_reload_time = 15
+        self.playlist_end: Optional[Sequence.num] = None
+        self.playlist_sequence: int = -1
+        self.playlist_sequences: List[Sequence] = []
+        self.playlist_reload_time: float = 6
         self.playlist_reload_time_override = self.session.options.get("hls-playlist-reload-time")
         self.playlist_reload_retries = self.session.options.get("hls-playlist-reload-attempts")
         self.live_edge = self.session.options.get("hls-live-edge")
@@ -190,30 +228,11 @@ class HLSStreamWorker(SegmentedStreamWorker):
         elif self.playlist_reload_time_override not in ["segment", "live-edge"]:
             self.playlist_reload_time_override = 0
 
-        self.reload_playlist()
-
-        if self.playlist_end is None:
-            if self.duration_offset_start > 0:
-                log.debug("Time offsets negative for live streams, skipping back {0} seconds",
-                          self.duration_offset_start)
-            # live playlist, force offset durations back to None
-            self.duration_offset_start = -self.duration_offset_start
-
-        if self.duration_offset_start != 0:
-            self.playlist_sequence = self.duration_to_sequence(self.duration_offset_start, self.playlist_sequences)
-
-        if self.playlist_sequences:
-            log.debug("First Sequence: {0}; Last Sequence: {1}",
-                      self.playlist_sequences[0].num, self.playlist_sequences[-1].num)
-            log.debug("Start offset: {0}; Duration: {1}; Start Sequence: {2}; End Sequence: {3}",
-                      self.duration_offset_start, self.duration_limit,
-                      self.playlist_sequence, self.playlist_end)
-
     def _reload_playlist(self, text, url):
-        return hls_playlist.load(text, url)
+        return load_hls_playlist(text, url)
 
     def reload_playlist(self):
-        if self.closed:
+        if self.closed:  # pragma: no cover
             return
 
         self.reader.buffer.wait_free()
@@ -228,11 +247,10 @@ class HLSStreamWorker(SegmentedStreamWorker):
             raise StreamError(err)
 
         if playlist.is_master:
-            raise StreamError("Attempted to play a variant playlist, use "
-                              "'hls://{0}' instead".format(self.stream.url))
+            raise StreamError(f"Attempted to play a variant playlist, use 'hls://{self.stream.url}' instead")
 
         if playlist.iframes_only:
-            raise StreamError("Streams containing I-frames only is not playable")
+            raise StreamError("Streams containing I-frames only are not playable")
 
         media_sequence = playlist.media_sequence or 0
         sequences = [Sequence(media_sequence + i, s)
@@ -243,12 +261,12 @@ class HLSStreamWorker(SegmentedStreamWorker):
         if sequences:
             self.process_sequences(playlist, sequences)
 
-    def _playlist_reload_time(self, playlist, sequences):
+    def _playlist_reload_time(self, playlist: M3U8, sequences: List[Sequence]) -> float:
         if self.playlist_reload_time_override == "segment" and sequences:
             return sequences[-1].segment.duration
         if self.playlist_reload_time_override == "live-edge" and sequences:
             return sum([s.segment.duration for s in sequences[-max(1, self.live_edge - 1):]])
-        if self.playlist_reload_time_override > 0:
+        if type(self.playlist_reload_time_override) is float and self.playlist_reload_time_override > 0:
             return self.playlist_reload_time_override
         if playlist.target_duration:
             return playlist.target_duration
@@ -257,7 +275,7 @@ class HLSStreamWorker(SegmentedStreamWorker):
 
         return self.playlist_reload_time
 
-    def process_sequences(self, playlist, sequences):
+    def process_sequences(self, playlist: M3U8, sequences: List[Sequence]) -> None:
         first_sequence, last_sequence = sequences[0], sequences[-1]
 
         if first_sequence.segment.key and first_sequence.segment.key.method != "NONE":
@@ -280,10 +298,11 @@ class HLSStreamWorker(SegmentedStreamWorker):
             else:
                 self.playlist_sequence = first_sequence.num
 
-    def valid_sequence(self, sequence):
+    def valid_sequence(self, sequence: Sequence) -> bool:
         return sequence.num >= self.playlist_sequence
 
-    def duration_to_sequence(self, duration, sequences):
+    @staticmethod
+    def duration_to_sequence(duration: int, sequences: List[Sequence]) -> int:
         d = 0
         default = -1
 
@@ -299,14 +318,38 @@ class HLSStreamWorker(SegmentedStreamWorker):
         return default
 
     def iter_segments(self):
+        try:
+            self.reload_playlist()
+        except StreamError as err:
+            log.error(f'{err}')
+            self.reader.close()
+            return
+
+        if self.playlist_end is None:
+            if self.duration_offset_start > 0:
+                log.debug(f"Time offsets negative for live streams, skipping back {self.duration_offset_start} seconds")
+            # live playlist, force offset durations back to None
+            self.duration_offset_start = -self.duration_offset_start
+
+        if self.duration_offset_start != 0:
+            self.playlist_sequence = self.duration_to_sequence(self.duration_offset_start, self.playlist_sequences)
+
+        if self.playlist_sequences:
+            log.debug(f"First Sequence: {self.playlist_sequences[0].num}; "
+                      f"Last Sequence: {self.playlist_sequences[-1].num}")
+            log.debug(f"Start offset: {self.duration_offset_start}; "
+                      f"Duration: {self.duration_limit}; "
+                      f"Start Sequence: {self.playlist_sequence}; "
+                      f"End Sequence: {self.playlist_end}")
+
         total_duration = 0
         while not self.closed:
             for sequence in filter(self.valid_sequence, self.playlist_sequences):
-                log.debug("Adding segment {0} to queue", sequence.num)
+                log.debug(f"Adding segment {sequence.num} to queue")
                 yield sequence
                 total_duration += sequence.segment.duration
                 if self.duration_limit and total_duration >= self.duration_limit:
-                    log.info("Stopping stream early after {0}".format(self.duration_limit))
+                    log.info(f"Stopping stream early after {self.duration_limit}")
                     return
 
                 # End of stream
@@ -320,29 +363,50 @@ class HLSStreamWorker(SegmentedStreamWorker):
                 try:
                     self.reload_playlist()
                 except StreamError as err:
-                    log.warning("Failed to reload playlist: {0}", err)
+                    log.warning(f"Failed to reload playlist: {err}")
 
 
 class HLSStreamReader(SegmentedStreamReader):
     __worker__ = HLSStreamWorker
     __writer__ = HLSStreamWriter
 
-    def __init__(self, stream, *args, **kwargs):
-        SegmentedStreamReader.__init__(self, stream, *args, **kwargs)
+    def __init__(self, stream):
         self.request_params = dict(stream.args)
-        self.timeout = stream.session.options.get("hls-timeout")
-
         # These params are reserved for internal use
         self.request_params.pop("exception", None)
         self.request_params.pop("stream", None)
         self.request_params.pop("timeout", None)
         self.request_params.pop("url", None)
 
+        self.filter_event = Event()
+        self.filter_event.set()
+
+        super().__init__(stream)
+
+    def read(self, size):
+        while True:
+            try:
+                return super().read(size)
+            except OSError:
+                # wait indefinitely until filtering ends
+                self.filter_event.wait()
+                if self.buffer.closed:
+                    return b""
+                # if data is available, try reading again
+                if self.buffer.length > 0:
+                    continue
+                # raise if not filtering and no data available
+                raise
+
+    def close(self):
+        super().close()
+        self.filter_event.set()
+
 
 class MuxedHLSStream(MuxedStream):
     __shortname__ = "hls-multi"
 
-    def __init__(self, session, video, audio, force_restart=False, ffmpeg_options=None, **args):
+    def __init__(self, session, video, audio, url_master=None, force_restart=False, ffmpeg_options=None, **args):
         tracks = [video]
         maps = ["0:v?", "0:a?"]
         if audio:
@@ -355,7 +419,11 @@ class MuxedHLSStream(MuxedStream):
         substreams = map(lambda url: HLSStream(session, url, force_restart=force_restart, **args), tracks)
         ffmpeg_options = ffmpeg_options or {}
 
-        super(MuxedHLSStream, self).__init__(session, *substreams, format="mpegts", maps=maps, **ffmpeg_options)
+        super().__init__(session, *substreams, format="mpegts", maps=maps, **ffmpeg_options)
+        self.url_master = url_master
+
+    def to_manifest_url(self):
+        return self.url_master
 
 
 class HLSStream(HTTPStream):
@@ -370,18 +438,23 @@ class HLSStream(HTTPStream):
     """
 
     __shortname__ = "hls"
+    __reader__ = HLSStreamReader
 
-    def __init__(self, session_, url, force_restart=False, start_offset=0, duration=None, **args):
-        HTTPStream.__init__(self, session_, url, **args)
+    def __init__(self, session_, url, url_master=None, force_restart=False, start_offset=0, duration=None, **args):
+        super().__init__(session_, url, **args)
+        self.url_master = url_master
         self.force_restart = force_restart
         self.start_offset = start_offset
         self.duration = duration
 
     def __repr__(self):
-        return "<HLSStream({0!r})>".format(self.url)
+        return f"<HLSStream({self.url!r}, {self.url_master!r})>"
 
     def __json__(self):
-        json = HTTPStream.__json__(self)
+        json = super().__json__()
+
+        if self.url_master:
+            json["master"] = self.url_master
 
         # Pretty sure HLS is GET only.
         del json["method"]
@@ -389,15 +462,18 @@ class HLSStream(HTTPStream):
 
         return json
 
+    def to_manifest_url(self):
+        return self.url_master
+
     def open(self):
-        reader = HLSStreamReader(self)
+        reader = self.__reader__(self)
         reader.open()
 
         return reader
 
     @classmethod
     def _get_variant_playlist(cls, res):
-        return hls_playlist.load(res.text, base_uri=res.url)
+        return load_hls_playlist(res.text, base_uri=res.url)
 
     @classmethod
     def parse_variant_playlist(cls, session_, url, name_key="name",
@@ -417,9 +493,6 @@ class HLSStream(HTTPStream):
                          name, pixels, bitrate.
         """
         locale = session_.localization
-        # Backwards compatibility with "namekey" and "nameprefix" params.
-        name_key = request_params.pop("namekey", name_key)
-        name_prefix = request_params.pop("nameprefix", name_prefix)
         audio_select = session_.options.get("hls-audio-select") or []
 
         res = session_.http.get(url, exception=IOError, **request_params)
@@ -427,7 +500,7 @@ class HLSStream(HTTPStream):
         try:
             parser = cls._get_variant_playlist(res)
         except ValueError as err:
-            raise IOError("Failed to parse playlist: {0}".format(err))
+            raise OSError("Failed to parse playlist: {0}".format(err))
 
         streams = OrderedDict()
         for playlist in filter(lambda p: not p.is_iframe, parser.playlists):
@@ -511,16 +584,16 @@ class HLSStream(HTTPStream):
             external_audio = preferred_audio or default_audio or fallback_audio
 
             if external_audio and FFMPEGMuxer.is_usable(session_):
-                external_audio_msg = u", ".join([
-                    u"(language={0}, name={1})".format(x.language, (x.name or "N/A"))
+                external_audio_msg = ", ".join([
+                    f"(language={x.language}, name={x.name or 'N/A'})"
                     for x in external_audio
                 ])
-                log.debug(u"Using external audio tracks for stream {0} {1}".format(
-                          stream_name, external_audio_msg))
+                log.debug(f"Using external audio tracks for stream {stream_name} {external_audio_msg}")
 
                 stream = MuxedHLSStream(session_,
                                         video=playlist.uri,
                                         audio=[x.uri for x in external_audio if x.uri],
+                                        url_master=url,
                                         force_restart=force_restart,
                                         start_offset=start_offset,
                                         duration=duration,
@@ -528,6 +601,7 @@ class HLSStream(HTTPStream):
             else:
                 stream = cls(session_,
                              playlist.uri,
+                             url_master=url,
                              force_restart=force_restart,
                              start_offset=start_offset,
                              duration=duration,
