@@ -7,7 +7,13 @@ from threading import Event, Thread
 import requests_mock
 
 from streamlink import Streamlink
-from streamlink.stream.hls import HLSStream
+from streamlink.stream.hls import HLSStream, HLSStreamWriter as _HLSStreamWriter
+
+
+TIMEOUT_AWAIT_READ = 5
+TIMEOUT_AWAIT_READ_ONCE = 5
+TIMEOUT_AWAIT_WRITE = 60  # https://github.com/streamlink/streamlink/issues/3868
+TIMEOUT_AWAIT_CLOSE = 5
 
 
 class HLSItemBase:
@@ -82,17 +88,46 @@ class Segment(HLSItemBase):
         )
 
 
+class EventedHLSStreamWriter(_HLSStreamWriter):
+    def __init__(self, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.write_wait = Event()
+        self.write_done = Event()
+        self.write_error = None
+
+    def _futures_put(self, item):
+        self.futures.put_nowait(item)
+
+    def _futures_get(self):
+        return self.futures.get_nowait()
+
+    @staticmethod
+    def _future_result(future):
+        return future.result(timeout=0)
+
+    def write(self, *args, **kwargs):
+        # only write once per step
+        self.write_wait.wait()
+        self.write_wait.clear()
+
+        try:
+            # don't write again during teardown
+            if not self.closed:
+                super().write(*args, **kwargs)
+        except Exception as err:
+            self.write_error = err
+            self.reader.close()
+        finally:
+            # notify main thread that writing has finished
+            self.write_done.set()
+
+
 class HLSStreamReadThread(Thread):
     """
     Run the reader on a separate thread, so that each read can be controlled from within the main thread
     """
-    def __init__(self, session, stream, *args, **kwargs):
-        """
-        :param Streamlink session:
-        :param HLSStream stream:
-        """
-        Thread.__init__(self, *args, **kwargs)
-        self.daemon = True
+    def __init__(self, session: Streamlink, stream: HLSStream, *args, **kwargs):
+        super().__init__(*args, **kwargs, daemon=True)
 
         self.read_wait = Event()
         self.read_once = Event()
@@ -103,12 +138,12 @@ class HLSStreamReadThread(Thread):
 
         self.session = session
         self.stream = stream
-        self.reader = stream.open()
+        self.reader = stream.__reader__(stream)
 
         # ensure that at least one read was attempted before closing the writer thread early
         # otherwise, the writer will close the reader's buffer, making it not block on read and yielding empty results
         def _await_read_then_close():
-            self.read_once.wait(timeout=5)
+            self.read_once.wait(timeout=TIMEOUT_AWAIT_READ_ONCE)
             return self.writer_close()
 
         self.writer_close = self.reader.writer.close
@@ -116,10 +151,11 @@ class HLSStreamReadThread(Thread):
 
     def run(self):
         while not self.reader.buffer.closed:
+            # at least one read was attempted
+            self.read_once.set()
             # only read once per step
             self.read_wait.wait()
             self.read_wait.clear()
-            self.read_once.set()
 
             try:
                 # don't read again during teardown
@@ -143,18 +179,24 @@ class HLSStreamReadThread(Thread):
         self.data[:] = []
         self.error = None
 
+    def close(self):
+        self.reader.buffer.close()
+        self.read_once.set()
+        self.read_wait.set()
+
 
 class TestMixinStreamHLS(unittest.TestCase):
     __stream__ = HLSStream
     __readthread__ = HLSStreamReadThread
 
+    session: Streamlink
+    stream: HLSStream
+    thread: HLSStreamReadThread
+
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.mocker = requests_mock.Mocker()
         self.mocks = {}
-        self.session = None
-        self.stream = None
-        self.thread = None
 
     def setUp(self):
         super().setUp()
@@ -162,54 +204,79 @@ class TestMixinStreamHLS(unittest.TestCase):
 
     def tearDown(self):
         super().tearDown()
-        self.close_thread()
+
+        # close read thread and make sure that all threads have terminated before moving on
+        self.close()
+        self.await_close()
+
         self.mocker.stop()
         self.mocks.clear()
-        self.session = None
-        self.stream = None
-        self.thread = None
 
     def mock(self, method, url, *args, **kwargs):
         self.mocks[url] = self.mocker.request(method, url, *args, **kwargs)
 
-    def called(self, item):
-        return self.mocks[self.url(item)].called
+    def get_mock(self, item):
+        return self.mocks[self.url(item)]
+
+    def called(self, item, once=False):
+        mock = self.get_mock(item)
+        return mock.called_once if once else mock.called
 
     def url(self, item):
         return item.url(self.id())
 
-    def content(self, segments, prop="content", cond=None):
-        return b"".join([getattr(segment, prop) for segment in segments.values() if cond is None or cond(segment)])
+    @staticmethod
+    def content(segments, prop="content", cond=None):
+        if isinstance(segments, dict):
+            segments = segments.values()
+        return b"".join([getattr(segment, prop) for segment in segments if cond is None or cond(segment)])
 
-    # close read thread and make sure that all threads have terminated before moving on
-    def close_thread(self):
+    def await_close(self, timeout=TIMEOUT_AWAIT_CLOSE):
         thread = self.thread
-        thread.reader.close()
-        thread.read_wait.set()
-        thread.reader.writer.join()
-        thread.reader.worker.join()
-        thread.join()
+        thread.reader.writer.join(timeout)
+        thread.reader.worker.join(timeout)
+        thread.join(timeout)
+
+    # make one write call on the write thread and wait until it has finished
+    def await_write(self, write_calls=1, timeout=TIMEOUT_AWAIT_WRITE):
+        writer = self.thread.reader.writer
+        if not writer.is_alive():  # pragma: no cover
+            raise RuntimeError("Write thread is not alive")
+
+        for write_call in range(write_calls):
+            writer.write_wait.set()
+            done = writer.write_done.wait(timeout)
+            if writer.write_error:
+                raise writer.write_error
+            if not done:  # pragma: no cover
+                raise RuntimeError(f"Await write timeout: write_call={write_call + 1}")
+            writer.write_done.clear()
 
     # make one read call on the read thread and wait until it has finished
-    def await_read(self, read_all=False, timeout=5):
+    def await_read(self, read_all=False, timeout=TIMEOUT_AWAIT_READ):
         thread = self.thread
+        if not thread.is_alive():  # pragma: no cover
+            raise RuntimeError("Read thread is not alive")
+
         thread.read_all = read_all
         thread.read_wait.set()
-        thread.read_done.wait(timeout)
-        thread.read_done.clear()
+        done = thread.read_done.wait(timeout)
 
         try:
             if thread.error:
                 raise thread.error
+            if not done:  # pragma: no cover
+                raise RuntimeError(f"Await read timeout: read_all={read_all}")
             return b"".join(thread.data)
         finally:
+            thread.read_done.clear()
             thread.reset()
 
     def get_session(self, options=None, *args, **kwargs):
         return Streamlink(options)
 
     # set up HLS responses, create the session and read thread and start it
-    def subject(self, playlists, options=None, streamoptions=None, threadoptions=None, *args, **kwargs):
+    def subject(self, playlists, options=None, streamoptions=None, threadoptions=None, start=True, *args, **kwargs):
         # filter out tags and duplicate segments between playlist responses while keeping index order
         segments_all = [item for playlist in playlists for item in playlist.items if isinstance(item, Segment)]
         segments = OrderedDict([(segment.num, segment) for segment in segments_all])
@@ -220,7 +287,20 @@ class TestMixinStreamHLS(unittest.TestCase):
 
         self.session = self.get_session(options, *args, **kwargs)
         self.stream = self.__stream__(self.session, self.url(playlists[0]), **(streamoptions or {}))
-        self.thread = self.__readthread__(self.session, self.stream, **(threadoptions or {}))
-        self.thread.start()
+        self.thread = self.__readthread__(self.session, self.stream, name=f"ReadThread-{self.id()}", **(threadoptions or {}))
+
+        if start:
+            self.start()
 
         return self.thread, segments
+
+    def start(self):
+        self.thread.reader.open()
+        self.thread.start()
+
+    def close(self):
+        thread = self.thread
+        thread.reader.close()
+        if isinstance(thread.reader.writer, EventedHLSStreamWriter):
+            thread.reader.writer.write_wait.set()
+        thread.read_wait.set()
